"""Primary CLI entrypoint."""

import typing
import logging

import click

from baymesh import node_validation, baymesh_versions
from baymesh.cli import devices, node_setup, echo

if typing.TYPE_CHECKING:
    pass

logger = logging.getLogger(__name__)


@click.group(no_args_is_help=True)
@click.pass_context
def cli(ctx: click.Context):
    """Node setup, validation, and management CLI for the Meshtastic Bay Area Group."""
    ctx.ensure_object(dict)
    try:
        is_up_to_date, latest_version = baymesh_versions.is_up_to_date()
    except RuntimeError as exc:
        logger.warning(f"Unable to determine latest version of baymesh: {exc}")
        return
    if not is_up_to_date:
        echo.warning(
            f"A new version ({latest_version}) of the baymesh CLI is available. "
            f"Please upgrade in order to ensure continued interop with the mesh."
        )


@cli.command()
def validate():
    """Validates that a connected node conforms to Baymesh standards."""
    port = devices.select_device_flow()
    interface = devices.ensure_meshtastic_interface(device_path=port.device, port=port)
    report = node_validation.validate_node(interface=interface)
    echo.info(
        f"Found Meshtastic node: { report.device_long_name } ({report.device_short_name})"
    )
    echo.working("Validating node configs...\n")
    node_validation.render_validation_report(report)


@cli.command()
def setup():
    """Sets up a node using the standard Baymesh configs."""
    port = devices.select_device_flow()
    interface = devices.ensure_meshtastic_interface(device_path=port.device, port=port)
    echo.confirm(
        "If you have already configured your node, the setup wizard may "
        "overwrite some of your settings. Continue?"
    )
    echo.working("Starting setup wizard...\n")
    setup_wizard = node_setup.SetupWizard()
    configs = setup_wizard.run()
    node_setup.apply_configs(configs, interface)
    echo.success("Device configured. Happy meshing!")


@cli.command()
def detect_devices():
    """Attempts to automatically detect supported devices.

    Only detects USB devices at the moment!
    """
    ports = devices.detect_supported_devices_via_serial()
    if not ports:
        click.echo("No supported devices found.")
        return
    click.echo("Found potentially supported devices:")
    for port in ports:
        click.echo(f"* {port.device}")
        click.echo(f"    Description: {port.description}")
        click.echo(f"           HWID: {port.hwid}")


if __name__ == "__main__":
    cli(obj={})



"""Checks to make sure that the user has the latest version of the CLI installed."""

import requests

from baymesh import __version__


def _version_from_pypi_url(url: str) -> str:
    """Given a PyPi source archive URL, return the package's version number."""
    source_archive_name = url.split("/")[-1]
    return source_archive_name.lstrip("baymesh-").rstrip(".tar.gz")


def get_latest_version() -> str:
    """Returns the latest version of the CLI available in PyPi."""
    try:
        response = requests.get("https://pypi.org/pypi/baymesh/json", timeout=3)
    except requests.exceptions.RequestException as exception:
        raise RuntimeError(f"Failed to get latest version from PyPi, {exception}")
    rdict = response.json()
    # Example JSON output: https://pypi.org/pypi/baymesh/json
    # urls keyword seems to be sorted and there is a wheel and a tarball per version.
    for url in rdict["urls"][-2:]:
        if url["url"].endswith(".tar.gz"):
            return _version_from_pypi_url(url["url"])
    raise RuntimeError("Unable to find the latest version of baymesh CLI.")


def get_current_version() -> str:
    """Returns the installed version of the CLI."""
    return __version__


def is_up_to_date() -> tuple[bool, str]:
    """Returns a Tuple of version info.

    Tuple is in the form of (<is_up_to_date>, <current_version>).
    """
    latest_version = get_latest_version()
    installed_version = get_current_version()
    up_to_date = latest_version == installed_version
    return up_to_date, latest_version



"""Logic for validating that nodes comply with Bay Mesh standards.

See:
* https://bayme.sh/docs/getting-started/recommended-settings/
* https://bayme.sh/docs/getting-started/mqtt/
"""

import enum
import dataclasses
import typing
from typing import Any

from baymesh.cli import echo
from meshtastic.protobuf import config_pb2, channel_pb2

from baymesh import firmware_versions

if typing.TYPE_CHECKING:
    import meshtastic.serial_interface


class RecommendationSeverity(enum.Enum):
    """Determines how much of an issue a report finding is."""

    # "Take it or leave it" advice.
    TIP = 1
    # "You should probably fix this, but it's not the end of the world if not".
    WARNING = 2
    # "You should *definitely* fix this unless you've got extenuating circumstances".
    ERROR = 3


@dataclasses.dataclass
class Recommendation(object):
    """Contains a single finding."""

    # The guidance that is provided to the user.
    message: str
    # A measure of how severe the issue is.
    severity: enum.Enum


@dataclasses.dataclass
class Report(object):
    """Contains a summarized report of findings for the user to consume."""

    # The device's long name, for display.
    device_long_name: str
    # The device's short name, also for display.
    device_short_name: str
    # A list of findings.
    recommendations: list[Recommendation] = dataclasses.field(default_factory=list)

    def add_recommendations(self, recommendations: list[Recommendation | None]):
        """Adds a recommendation to the report."""
        for recommendation in recommendations:
            if not recommendation:
                continue
            self.recommendations += (recommendation,)

    def list_recommendations(self, only_severity=None) -> list[Recommendation]:
        """Lists (and optionally filters) the report's recommendations."""
        if not only_severity:
            return self.recommendations
        matches = []
        for recommendation in self.recommendations:
            if recommendation.severity == only_severity:
                matches.append(recommendation)
        return matches

    def validation_successful(self) -> bool:
        """Returns True if the node complies with all essential standards."""
        errors = self.list_recommendations(only_severity=RecommendationSeverity.ERROR)
        return len(errors) == 0


def validate_node(interface: "meshtastic.serial_interface.SerialInterface") -> Report:
    """Given a serial path, validate the attached device for compliance."""
    node_info: dict[str, Any] | None = interface.getMyNodeInfo()
    if not node_info:
        raise RuntimeError("Failed to get node info. Please try again.")
    our_node: meshtastic.Node = interface.getNode("^local")

    report = Report(
        device_long_name=node_info["user"]["longName"],
        device_short_name=node_info["user"]["shortName"],
    )
    # Run through all of the validation functions for the things that we care about.
    # Recommendations will be shown to the user in the order that the checking functions
    # appear below.
    report.add_recommendations(
        [
            _check_user_long_name(long_name=node_info["user"]["longName"]),
            _check_lora_region(region=our_node.localConfig.lora.region),
            _check_lora_preset(
                use_preset=our_node.localConfig.lora.use_preset,
                modem_preset=our_node.localConfig.lora.modem_preset,
            ),
            _check_lora_hop_limit(hop_limit=our_node.localConfig.lora.hop_limit),
            _check_device_role(role=our_node.localConfig.device.role),
            _check_module_telemetry_update_interval(
                telemetry_type="device",
                update_interval=our_node.moduleConfig.telemetry.device_update_interval,
            ),
            _check_module_telemetry_update_interval(
                telemetry_type="environment",
                update_interval=our_node.moduleConfig.telemetry.environment_update_interval,
            ),
            _check_module_telemetry_update_interval(
                telemetry_type="power",
                update_interval=our_node.moduleConfig.telemetry.power_update_interval,
            ),
            _check_node_firmware_version(
                node_firmware_version=firmware_versions.get_node_firmware_version(
                    our_node
                ),
                latest_firmware_version=firmware_versions.get_latest_firmware_version(),
            ),
            _check_channels_mqtt_settings(
                channels=our_node.channels,
            ),
        ]
    )

    interface.close()
    return report


def _check_user_long_name(long_name: str) -> Recommendation | None:
    """Suggest that the user set a long name if it looks like they've got the default."""
    split_name = long_name.split()
    if len(split_name) == 2 and split_name[0] == "Meshtastic" and len(split_name[1]) == 4:
        return Recommendation(
            message=f"Your node's long name '{long_name}' is close to the default. "
            f"Consider setting a more descriptive long name.",
            severity=RecommendationSeverity.WARNING,
        )


def _check_lora_region(
    region: config_pb2.Config.LoRaConfig.RegionCode.ValueType,
) -> Recommendation | None:
    """Ensure that the user has selected the correct region."""
    if not region == config_pb2.Config.LoRaConfig.RegionCode.US:
        region_name = config_pb2.Config.LoRaConfig.RegionCode.Name(region)
        return Recommendation(
            message=f"Region is set to {region_name} instead of US.",
            severity=RecommendationSeverity.ERROR,
        )


def _check_lora_preset(use_preset: bool, modem_preset: int) -> Recommendation | None:
    if (
        not use_preset
        or modem_preset != config_pb2.Config.LoRaConfig.ModemPreset.MEDIUM_SLOW
    ):
        """Ensure that the user is using the MEDIUM_SLOW preset."""
        return Recommendation(
            message="Lora preset is not set to MEDIUM_SLOW.",
            severity=RecommendationSeverity.ERROR,
        )


def _check_lora_hop_limit(hop_limit: int) -> Recommendation | None:
    """Suggest that the user set our suggested hop limit."""
    if hop_limit != 3:
        return Recommendation(
            message=f"We recommend a hop limit of 3 vs your current value of {hop_limit}.",
            severity=RecommendationSeverity.WARNING,
        )


def _check_device_role(role: int) -> Recommendation | None:
    """Ensure that the user isn't using a node role that we discourage."""
    discouraged_roles = [
        config_pb2.Config.DeviceConfig.Role.ROUTER,
        config_pb2.Config.DeviceConfig.Role.ROUTER_CLIENT,
        config_pb2.Config.DeviceConfig.Role.REPEATER,
    ]
    if role in discouraged_roles:
        role_name = config_pb2.Config.DeviceConfig.Role.Name(role)
        return Recommendation(
            message=f"Please do not use the {role_name} node role. "
            "Consider CLIENT or CLIENT_MUTE instead.",
            severity=RecommendationSeverity.ERROR,
        )


def _check_module_telemetry_update_interval(
    telemetry_type: str,
    update_interval: int,
) -> Recommendation | None:
    """Suggest that the user keep telemetry intervals to a reasonable level."""
    if update_interval and update_interval < 3600:
        return Recommendation(
            message=f"To help us reduce mesh congestion, lower your {telemetry_type} "
            f"telemetry update interval from {update_interval} to 3600 seconds.",
            severity=RecommendationSeverity.WARNING,
        )


def _check_node_firmware_version(
    node_firmware_version: "firmware_versions.FirmwareVersion",
    latest_firmware_version: "firmware_versions.FirmwareVersion",
) -> Recommendation | None:
    """Ensure that the user is on a current firmware version."""
    if node_firmware_version < latest_firmware_version:
        return Recommendation(
            message=f"To ensure a healthy mesh, please upgrade to the latest"
            f"Meshtastic firmware version ({latest_firmware_version}).",
            severity=RecommendationSeverity.ERROR,
        )


def _check_channels_mqtt_settings(
    channels: list[channel_pb2.Channel] | None,
) -> Recommendation | None:
    """Ensure that the user has not set the dreaded downlink_enable to True."""
    if not channels:
        # Unclear on when this won't be set, but apparently it's not a given!
        return
    for channel in channels:
        if not channel.settings.downlink_enabled:
            return
        channel_index = channel.index
        channel_name = channel_pb2.Channel.Role.Name(channel.role)
        return Recommendation(
            message=f"Disable MQTT uplink on channel {channel_index}: {channel_name}.",
            severity=RecommendationSeverity.ERROR,
        )


def render_validation_report(report: "Report"):
    """Renders the validation report for consumption by the user."""
    success_msg = "Your node is compliant with all Meshtastic Bay Area Group standards!"
    if not report.list_recommendations():
        echo.success(success_msg)
        return

    for recommendation in report.recommendations:
        echo_func = echo._recommendation_severity_to_echo(recommendation.severity)
        echo_func(f"{ recommendation.severity.name }: { recommendation.message }")

    if report.validation_successful():
        echo.success(f"{ success_msg } Please consider the above warning(s).")
    else:
        echo.error(
            "Your node is not complaint with Meshtastic Bay Area Group standards "
            "due to the above error(s)."
        )



"""Node setup wizard logic."""

import logging
import sys
import time
import typing
import dataclasses

import click

from baymesh.cli import devices, echo
from meshtastic.protobuf import config_pb2

if typing.TYPE_CHECKING:
    from typing import Callable
    import meshtastic.serial_interface

    LoraPresetValueType = config_pb2.Config.LoRaConfig.ModemPreset.ValueType
    DeviceRoleValueType = config_pb2.Config.DeviceConfig.Role.ValueType


class SetupWizard(object):
    """Contains a guided flow for collecting settings from users."""

    def __init__(self):
        self._settings: "ConfigsToApply" = ConfigsToApply()
        self._current_prompt = None
        self._is_active = False

    def run(self) -> "ConfigsToApply":
        """Begin the interactive guided flow and return the configs when done."""
        self._settings = ConfigsToApply()
        self._current_prompt = self._prompt_lora_preset
        self._is_active = True
        while self._is_active:
            self._current_prompt()
        return self._settings

    def _set_prompt_func(self, prompt: "Callable"):
        """Set the prompt function to call on the next self.run() iteration."""
        self._current_prompt = prompt

    def _finish(self):
        """Complete the process of gathering settings from the user."""
        # Setting this to False causes the while loop in self.run() to fall through.
        self._is_active = False

    def _prompt_lora_preset(self):
        use_medium_slow = echo.confirm(
            message="Use MEDIUM_SLOW LoRa preset?",
            additional_info="Start with MEDIUM_SLOW. Return and answer No if you can't find other nodes.",
            default=True,
        )
        if use_medium_slow:
            self._settings.lora_preset = (
                config_pb2.Config.LoRaConfig.ModemPreset.MEDIUM_SLOW
            )
        else:
            self._settings.lora_preset = (
                config_pb2.Config.LoRaConfig.ModemPreset.LONG_FAST
            )
        self._set_prompt_func(self._prompt_long_name)

    def _prompt_long_name(self):
        long_name = echo.prompt(
            message="What would you like this node's long name to be?",
            additional_info="This is typically your name, Discord handle, or Ham callsign",
        )
        self._settings.device_long_name = long_name
        self._set_prompt_func(self._prompt_short_name)

    def _prompt_short_name(self):
        short_name = echo.prompt(
            message="What would you like this node's short name to be?",
            additional_info="This can be up to four characters long and can even be an emoji",
        )
        if len(short_name) > 4:
            echo.error("Short name must be no longer than four characters.")
            return
        self._settings.device_short_name = short_name
        self._set_prompt_func(self._prompt_device_role)

    def _prompt_device_role(self):
        is_relay = echo.confirm(
            message="Should this node relay messages from other nodes to other nodes in the mesh?",
            additional_info=[
                "If this is your primary node or a home base station, answer Y(yes).",
                "If this node will be in close proximity to other nodes who are already relaying messages, answer N(no)",
            ],
            default=True,
        )
        if is_relay:
            device_role = config_pb2.Config.DeviceConfig.Role.CLIENT
        else:
            device_role = config_pb2.Config.DeviceConfig.Role.CLIENT_MUTE
        self._settings.device_role = device_role
        self._set_prompt_func(self._prompt_confirm_settings)

    def _prompt_confirm_settings(self):
        click.secho("\nSettings preview: not yet applied to device", bold=True)
        click.echo(self._settings)
        is_satisfied = echo.confirm("Apply these configs to your device?", default=True)
        if is_satisfied:
            self._finish()
        else:
            click.echo("Canceling setup and exiting.")
            sys.exit(0)


@dataclasses.dataclass
class ConfigsToApply(object):
    """Simplified local configs container for the setup wizard."""

    device_long_name: str = ""
    device_short_name: str = ""

    lora_preset: "LoraPresetValueType" = (
        config_pb2.Config.LoRaConfig.ModemPreset.MEDIUM_SLOW
    )
    device_role: "DeviceRoleValueType" = config_pb2.Config.DeviceConfig.Role.CLIENT

    telemetry_device_update_interval: int = 3600
    telemetry_environment_update_interval: int = 3600
    telemetry_power_update_interval: int = 3600

    @property
    def lora_preset_name(self) -> str:
        """String form of the selected LoRa preset."""
        return config_pb2.Config.LoRaConfig.ModemPreset.Name(self.lora_preset)

    @property
    def device_role_name(self) -> str:
        """String form of the selected device role."""
        return config_pb2.Config.DeviceConfig.Role.Name(self.device_role)

    def __str__(self):
        return (
            f"Long name: {self.device_long_name}\n"
            f"Short name: {self.device_short_name}\n"
            f"LoRa preset: {self.lora_preset_name}\n"
            f"Role: {self.device_role_name}\n"
        )


def apply_configs(
    configs: ConfigsToApply, interface: "meshtastic.serial_interface.SerialInterface"
):
    """Applies the configs based on the users responses to the setup wizard."""
    our_node: meshtastic.Node = interface.getNode("^local")

    # Meshtastic spews some warnings that we don't care about during this process.
    original_level = echo.get_logging_level()
    echo.set_logging_level(logging.ERROR)

    click.secho("⚙️  Applying long and short name...")
    # This doesn't seem to reboot the node...
    our_node.setOwner(
        long_name=configs.device_long_name, short_name=configs.device_short_name
    )

    # ... but we'll give it a second to settle, anyway.
    time.sleep(2)

    click.secho("⚙️  Applying LoRa configs...")
    our_node.localConfig.lora.use_preset = True
    our_node.localConfig.lora.region = config_pb2.Config.LoRaConfig.RegionCode.US
    our_node.localConfig.lora.modem_preset = configs.lora_preset
    our_node.localConfig.lora.config_ok_to_mqtt = True
    our_node.writeConfig("lora")

    time.sleep(4)
    interface, our_node = devices.block_until_device_returns_node(interface)

    click.secho("⚙️  Applying role configs...")
    our_node.localConfig.device.role = configs.device_role
    our_node.writeConfig("device")

    time.sleep(4)
    # Restore logging level.
    echo.set_logging_level(original_level)



"""Logic for echoing with formatting in a consistent manner."""

import enum
import typing
import logging

import click

from baymesh import node_validation

if typing.TYPE_CHECKING:
    from typing import Callable, Any


ERROR_COLOR = "red"
SUCCESS_COLOR = "green"
WARNING_COLOR = "yellow"
INFO_COLOR = "cyan"


def error(message):
    """Indicate that an error has occurred."""
    click.secho(f"🚨 {message}", fg=ERROR_COLOR, bold=True)


def success(message):
    """Indicate that a success has occurred."""
    click.secho(f"✅ {message}", fg=SUCCESS_COLOR, bold=True)


def warning(message):
    """Non-blocking warning."""
    click.secho(f"⚠️  {message}", fg=WARNING_COLOR)


def info(message):
    """Share context or progress."""
    click.secho(f"️ℹ️  {message}", fg=INFO_COLOR)


def working(message):
    """Share progress."""
    click.secho(f"⚙️  {message}")


def _buffer_additional_info(additional_info: str | list[str]) -> str:
    """Convenience function for formatting additional info beneath a prompt."""
    if not additional_info:
        return ""
    if isinstance(additional_info, str):
        additional_info = [additional_info]
    buf = ""
    for additional_info_nugget in additional_info:
        buf += f"\n     {additional_info_nugget}"
    return buf


def confirm(message: str, additional_info: str | list[str] = "", **kwargs) -> bool:
    """Prompts the user for a confirmation."""
    msg_contents = click.style(f"🤔 {message}", fg=INFO_COLOR)
    msg_contents += _buffer_additional_info(additional_info)
    return click.confirm(msg_contents, **kwargs)


def prompt(message: str, additional_info: str | list[str] = "", **kwargs) -> "Any":
    """Prompts the user."""
    msg_contents = click.style(f"🤔 {message}", fg=INFO_COLOR)
    msg_contents += _buffer_additional_info(additional_info)
    return click.prompt(msg_contents, **kwargs)


def get_logging_level(logger=None) -> int:
    """Gets the current level for the given logger."""
    logger = logging.getLogger(logger)
    return logger.getEffectiveLevel()


def set_logging_level(level: int, logger=None):
    """Sets the specified logger to the given level."""
    logger = logging.getLogger(logger)
    logger.setLevel(level)


def _recommendation_severity_to_echo(severity: "enum.Enum") -> "Callable":
    """Maps a recommendation severity to the corresponding echo func."""
    match severity:
        case node_validation.RecommendationSeverity.ERROR:
            return error
        case node_validation.RecommendationSeverity.WARNING:
            return warning
        case _:
            return info



"""Firmware version munging."""

import io
import typing
import functools
import dataclasses
import contextlib

import requests

if typing.TYPE_CHECKING:
    import meshtastic  # pyright: ignore [reportMissingTypeStubs]


@functools.total_ordering
@dataclasses.dataclass
class FirmwareVersion(object):
    """Internal representation of a Meshtastic firmware version."""

    major: int
    minor: int
    patch: int
    commit_hash: str

    def __str__(self):
        return f"{self.major}.{self.minor}.{self.patch}.{self.commit_hash}"

    def __eq__(self, other) -> bool:
        if not self._is_valid_operand(other):
            return NotImplemented
        return self.as_tuple() == other.as_tuple()

    def __gt__(self, other) -> bool:
        if not self._is_valid_operand(other):
            return NotImplemented
        if self.major > other.major:
            return True
        if self.major >= other.major and self.minor > other.minor:
            return True
        if (
            self.major >= other.major
            and self.minor >= other.minor
            and self.patch >= other.patch
        ):
            return True
        return False

    @staticmethod
    def _is_valid_operand(other) -> bool:
        """Determines whether the other object is comparable."""
        return (
            hasattr(other, "major")
            and hasattr(other, "minor")
            and hasattr(other, "patch")
            and callable(getattr(other, "as_tuple"))
        )

    def as_tuple(self) -> "typing.Tuple[int, int, int, str]":
        """Returns a tuple of all version elements."""
        return self.major, self.minor, self.patch, self.commit_hash


def _parse_firmware_version_str(firmware_str: str) -> FirmwareVersion:
    """Parses a Meshtastic firmware version string and returns a FirmwareVersion."""
    version_split: list[str] = firmware_str.split(".")
    return FirmwareVersion(
        major=int(version_split[0]),
        minor=int(version_split[1]),
        patch=int(version_split[2]),
        commit_hash=version_split[3],
    )


def get_node_firmware_version(node: "meshtastic.Node") -> FirmwareVersion:
    """Retrieves the connected node's firmware version and returns a FirmwareVersion.

    The meshtastic library's Node.getMetadata() function prints out the firmware version
    instead of returning it, so we've got to work around by redirecting stdout.
    """
    output_buf = io.StringIO()
    with contextlib.redirect_stdout(output_buf):
        node.getMetadata()
    output = output_buf.getvalue()
    version_str = output.split("firmware_version: ")[1].split("\n")[0]
    return _parse_firmware_version_str(version_str)


def get_latest_firmware_version() -> FirmwareVersion:
    """Retrieves the latest firmware version from GitHub and returns a FirmwareVersion."""
    response = requests.get(
        "https://api.github.com/repos/meshtastic/firmware/releases/latest"
    )
    release = response.json()
    tag_name = release["tag_name"]
    # The upstream git tags have a "v" prefix that we need to remove for consistency.
    version_str = tag_name.lstrip("v")
    return _parse_firmware_version_str(version_str)



from meshtastic.protobuf import config_pb2, channel_pb2

from baymesh import node_validation, firmware_versions


class TestNodeValidationChecks(object):
    """Test suite for node validation checks."""

    @staticmethod
    def _assert_success(
        recommendation: node_validation.Recommendation | None, message=None
    ):
        """Checks for a successful validation run."""
        assert recommendation is None, (
            message or "Expected no recommendation but found one."
        )

    @staticmethod
    def _assert_recommendation(
        recommendation: node_validation.Recommendation | None,
        severity: node_validation.RecommendationSeverity,
        message=None,
    ):
        """Checks for a successful validation run."""
        assert recommendation is not None, (
            message or "Expected a recommendation but found none."
        )
        if severity:
            assert recommendation.severity == severity, (
                message or "Severity does not match expected value."
            )

    def test_check_user_long_name(self):
        """Exercise node long name validation."""
        assert node_validation._check_user_long_name("This should pass") is None
        # Error if the device has a name that looks like a default.
        assert node_validation._check_user_long_name("Meshtastic 3cc8") is not None

    def test_check_lora_region(self):
        """Exercise node region validation."""
        self._assert_success(
            node_validation._check_lora_region(
                region=config_pb2.Config.LoRaConfig.RegionCode.US
            )
        )
        self._assert_recommendation(
            node_validation._check_lora_region(
                region=config_pb2.Config.LoRaConfig.RegionCode.EU_433
            ),
            severity=node_validation.RecommendationSeverity.ERROR,
        )

    def test_check_lora_preset(self):
        """Exercise node lora preset validation."""
        self._assert_success(
            node_validation._check_lora_preset(
                use_preset=True,
                modem_preset=config_pb2.Config.LoRaConfig.ModemPreset.MEDIUM_SLOW,
            )
        )
        self._assert_recommendation(
            node_validation._check_lora_preset(
                use_preset=True,
                modem_preset=config_pb2.Config.LoRaConfig.ModemPreset.LONG_FAST,
            ),
            severity=node_validation.RecommendationSeverity.ERROR,
        )
        self._assert_recommendation(
            node_validation._check_lora_preset(
                use_preset=False,
                modem_preset=config_pb2.Config.LoRaConfig.ModemPreset.MEDIUM_SLOW,
            ),
            severity=node_validation.RecommendationSeverity.ERROR,
        )

    def test_check_lora_hop_limit(self):
        """Exercise node lora hop limit validation."""
        self._assert_success(node_validation._check_lora_hop_limit(3))
        self._assert_recommendation(
            node_validation._check_lora_hop_limit(2),
            severity=node_validation.RecommendationSeverity.WARNING,
        )
        self._assert_recommendation(
            node_validation._check_lora_hop_limit(4),
            severity=node_validation.RecommendationSeverity.WARNING,
        )

    def test_check_device_role(self):
        """Exercise node device role validation."""
        self._assert_success(
            node_validation._check_device_role(config_pb2.Config.DeviceConfig.Role.CLIENT)
        )
        self._assert_success(
            node_validation._check_device_role(
                config_pb2.Config.DeviceConfig.Role.CLIENT_MUTE
            )
        )
        self._assert_success(
            node_validation._check_device_role(
                config_pb2.Config.DeviceConfig.Role.CLIENT_HIDDEN
            )
        )
        self._assert_success(
            node_validation._check_device_role(
                config_pb2.Config.DeviceConfig.Role.ROUTER_LATE
            )
        )
        self._assert_recommendation(
            node_validation._check_device_role(
                config_pb2.Config.DeviceConfig.Role.ROUTER
            ),
            severity=node_validation.RecommendationSeverity.ERROR,
        )

    def test_check_module_telemetry_update_interval(self):
        """Exercise node module telemetry update intervals validation."""
        self._assert_success(
            node_validation._check_module_telemetry_update_interval(
                telemetry_type="device",
                update_interval=60 * 60,
            )
        )
        self._assert_recommendation(
            node_validation._check_module_telemetry_update_interval(
                telemetry_type="device",
                update_interval=60 * 30,
            ),
            severity=node_validation.RecommendationSeverity.WARNING,
        )

    def test_check_node_firmware_version(self):
        """Exercise node firmware version validation."""

        newer = firmware_versions.FirmwareVersion(
            major=2, minor=5, patch=18, commit_hash="89ebafc"
        )
        older = firmware_versions.FirmwareVersion(
            major=2, minor=5, patch=15, commit_hash="79da236"
        )
        self._assert_success(
            node_validation._check_node_firmware_version(
                node_firmware_version=newer,
                latest_firmware_version=older,
            )
        )
        self._assert_success(
            node_validation._check_node_firmware_version(
                node_firmware_version=newer,
                latest_firmware_version=newer,
            )
        )
        self._assert_recommendation(
            node_validation._check_node_firmware_version(
                node_firmware_version=older,
                latest_firmware_version=newer,
            ),
            severity=node_validation.RecommendationSeverity.ERROR,
        )

    def test_check_channels_mqtt_settings(self):
        """Exercise node channels mqtt setting validation."""

        valid_channel_0 = channel_pb2.Channel(
            index=0,
            settings=channel_pb2.ChannelSettings(
                downlink_enabled=False,
            ),
            role=channel_pb2.Channel.Role.PRIMARY,
        )
        valid_channel_list = [valid_channel_0]
        self._assert_success(
            node_validation._check_channels_mqtt_settings(channels=valid_channel_list)
        )

        invalid_channel_0 = channel_pb2.Channel(
            index=0,
            settings=channel_pb2.ChannelSettings(
                downlink_enabled=True,
            ),
            role=channel_pb2.Channel.Role.PRIMARY,
        )
        invalid_channel_list = [invalid_channel_0]
        self._assert_recommendation(
            node_validation._check_channels_mqtt_settings(channels=invalid_channel_list),
            severity=node_validation.RecommendationSeverity.ERROR,
        )



"""Device connection management logic."""

import sys
import time
import typing

import click
import meshtastic
import meshtastic.serial_interface
import serial
import serial.tools.list_ports

from baymesh.cli import echo

if typing.TYPE_CHECKING:
    from serial.tools.list_ports_common import ListPortInfo

    IFNPair = typing.Tuple[meshtastic.serial_interface.SerialInterface, "meshtastic.Node"]


def detect_supported_devices_via_serial() -> list["ListPortInfo"]:
    """Returns a list of ports that may have a supported device connected."""
    supported_devices = meshtastic.util.detect_supported_devices()
    if not supported_devices:
        return []
    candidate_ports = meshtastic.util.active_ports_on_supported_devices(supported_devices)
    all_serial_ports = serial.tools.list_ports.comports()
    matches = []
    for serial_port in all_serial_ports:
        if serial_port.device in candidate_ports:
            matches.append(serial_port)
    return matches


def select_device_flow() -> "ListPortInfo":
    """If multiple supported devices are connected, prompt the user to select one."""
    ports = detect_supported_devices_via_serial()
    if not ports:
        raise click.UsageError("Could not find a supported device connected via USB.")
    if len(ports) == 1:
        # Only one device, no need to prompt!
        return ports[0]

    while True:
        click.echo("Found multiple supported devices:")
        for i, port in enumerate(ports):
            click.echo(f"{i}) {port.device} - {port.description} - {port.hwid}")
        num_selected = int(
            click.prompt(
                "Which device number would you like to validate?",
                type=click.Choice(list(str(c) for c in range(len(ports)))),
            )
        )
        return ports[num_selected]


def ensure_meshtastic_interface(
    device_path: str, port: "ListPortInfo"
) -> meshtastic.serial_interface.SerialInterface:
    """Given a device path, returns a SerialInterface or non-zero exits."""
    click.echo(f"⚙️  Opening connection to {port.description} via {port.device}...")
    try:
        return meshtastic.serial_interface.SerialInterface(device_path)
    except OSError as e:
        if "Resource busy" in str(e):
            echo.error(
                "Device is busy. If you have another configuration client open, "
                "please close it and try again."
            )
            sys.exit(1)
        raise


def block_until_device_returns(
    interface: "meshtastic.serial_interface.SerialInterface",
) -> "meshtastic.serial_interface.SerialInterface":
    """Block until the specified devices becomes responsive.

    Most useful for applying configs and waiting for the device to come back up.
    """
    while True:
        try:
            # Just trying to trigger an exception if the device isn't ready.
            interface = meshtastic.serial_interface.SerialInterface(interface.devPath)
            # Give it another couple of seconds to settle up, since we were seeing
            # warnings with some devices.
            time.sleep(2)
            return interface
        except serial.SerialException:
            time.sleep(1)


def block_until_device_returns_node(
    interface: "meshtastic.serial_interface.SerialInterface",
) -> "IFNPair":
    """Same as block_until_device_returns but also returns a meshtastic.Node."""
    our_node = block_until_device_returns(interface)
    return interface, our_node.getNode("^local")
